#include <stddef.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>

#include "modsetup.h"
#include "xscommon.h"
#include "keyboard.h"
#include "jediedit.h"

#define DF_FALSE (0x00000000UL)
#define DF_TRUE  (0xFFFFFFFFUL)

#pragma pack(push, 1)
typedef struct {
  BYTE kb[25]; /* keyboard */
  BYTE js[16]; /* joystick */
  BYTE ms[16]; /* mouse */
} cfg_ctrl;

typedef struct {
  BYTE  mark;              /* 0xCF - ConFig? */
  DWORD bHeadWave;         /* 0 / -1 */
  DWORD bShowHUD;          /* 0 / -1 */
  DWORD iGraphDetail;      /* [0..2] --- [Max, Mid, Min] */
  DWORD iShipDetail;       /* [0..3] --- [All, Max, Mid, Min] */
  DWORD iScreenSize;       /* [0..9] */
  BYTE  bAllowKB;          /* 0 / -1 */
  BYTE  bAllowMS;          /* 0 / -1 */
  BYTE  bAllowJS;          /* 0 / -1 */
  DWORD iGameVolSound;     /* [0..9] => 0x00..0x7F */
  DWORD iGameVolMusic;     /* [0..9] => 0x00..0x7F */
  DWORD iMouseSensitivity; /* [0..9] (0 - disabled; 4 - enabled in SETUP.EXE) */
  DWORD iGammaCorrection;  /* [0..4] [Off, Min, Mid, Max] */
  DWORD iSuperShield;      /* 0 / 1 - this can be 1 instead of -1 */
  DWORD iChannelsNumber;   /* [4..8] */
  DWORD bWeaponAutoMount;  /* 0 / -1 */
  DWORD iCutsVolSound;     /* [0..9] => 0x00..0x7F */
  DWORD iCutsVolMusic;     /* [0..9] => 0x00..0x7F --- (127*l) / 9 */
} cfg_main;

typedef struct {
  BYTE  mark;    /* 0xDB - DataBase? */
  DWORD list[9]; /* -1, -1, -1, -1, -1, -1, 0, 0, 0 */
} cfg_data;

typedef struct {
  char     mark[3];  /* CFG */
  BYTE     vers;     /* 100 (1.00 ?) */
  DWORD    len1;     /* =60 - sizeof(main) */
  DWORD    len2;     /* =37 - sizeof(data) */
  DWORD    len3;     /* =84 - sizeof(list) */
  DWORD    len4;     /* =57 - sizeof(ctrl) */
  cfg_main main;     /* (60) game configuration */
  cfg_data data;     /* (37) unknown, entirely skipped by the game on load */
  DWORD    list[21]; /* (84) additional values for joystick */
  cfg_ctrl ctrl;     /* (57) input configuration */
} cfg_jedi;
#pragma pack(pop)

static cfg_jedi dfconfig;

/* keyboard */
static BYTE kbidx[] = {2, 3, 11, 10, 9, 8, 14, 12, 13, 6, 7, 16, 4, 5, 21, 22, 15};
/* to, from */
static BYTE kbmap[][2] = {{0, 10}, {24, 10}, {1, 11}, {23, 11}, {20, 2}, {19, 3}};
/* mouse */
static BYTE msidx[] = {0, 1, 2, 3, 7, 5, 6, 11, 10/*v1.4*/, 9, 12, 13, 8};
/* joystick */
static BYTE jsidx[] = {2, 1, 3, 6, 4, 5, 10, 11, 8, 12, 13, 14, 7};
static BYTE jsmap[][2] = {{0, 2}};
/* (j)oy(s)tick types (to) (f)ile from menu */
static BYTE jstof[] = {0, 1, 6, 7, 2, 3};
/* (j)oy(s)tick types (to) (m)enu from file */
static BYTE jstom[] = {0, 1, 4, 5, 0, 0, 2, 3};

/* default configuration file contents */
static BYTE defcfg[] = {
  0x43, 0x46, 0x47, 0x64, 0x3C, 0x00, 0x00, 0x00, 0x25, 0x00, 0x00, 0x00, 0x54, 0x00, 0x00, 0x00,
  0x39, 0x00, 0x00, 0x00, 0xCF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x00, 0x7F, 0x00, 0x00, 0x00,
  0x64, 0x00, 0x00, 0x00, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x7F, 0x00, 0x00, 0x00, 0x7F, 0x00, 0x00, 0x00,
  0xDB, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x4D, 0x4B, 0x48, 0x50, 0x2A, 0x3A, 0x1D,
  0x2C, 0x34, 0x33, 0x4D, 0x4B, 0x2E, 0x39, 0x2D, 0x4C, 0x38, 0x00, 0x00, 0x50, 0x48, 0x51, 0x49,
  0x4B, 0x4D, 0x01, 0x02, 0x01, 0x0F, 0x0A, 0x04, 0x0E, 0x00, 0x04, 0x00, 0x03, 0x05, 0x00, 0x00,
  0x10, 0x00, 0x01, 0x04, 0x06, 0x00, 0x03, 0x00, 0x00, 0x00, 0x00, 0x07, 0x00, 0x05, 0x00, 0x00,
  0x00, 0x00
};

/* v1.4 - modern control configuration */
static BYTE mdrcfg[] = {
  0x43, 0x46, 0x47, 0x64, 0x3C, 0x00, 0x00, 0x00, 0x25, 0x00, 0x00, 0x00, 0x54, 0x00, 0x00, 0x00,
  0x39, 0x00, 0x00, 0x00, 0xCF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x00, 0x7F, 0x00, 0x00, 0x00,
  0x64, 0x00, 0x00, 0x00, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x08, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x7F, 0x00, 0x00, 0x00, 0x7F, 0x00, 0x00, 0x00,
  0xDB, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x11, 0x1F, 0x2A, 0x00, 0x00,
  0x00, 0x20, 0x1E, 0x00, 0x00, 0x1D, 0x12, 0x39, 0x00, 0x00, 0x00, 0x00, 0x1F, 0x11, 0x00, 0x00,
  0x00, 0x00, 0x01, 0x02, 0x01, 0x0F, 0x0A, 0x04, 0x0E, 0x00, 0x04, 0x00, 0x03, 0x05, 0x00, 0x00,
  0x10, 0x00, 0x01, 0x04, 0x00, 0x00, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x05, 0x00, 0x00,
  0x00, 0x00
};

static char skb_bind[(32 + 1) * 17];
static menuitem kb_bind[] = {
  {"Keyboard setup", NULL, 17},
  {"Forward", NULL, 0},
  {"Backward", NULL, 0},
  {"Turn Left", NULL, 0},
  {"Turn Right", NULL, 0},
  {"Strafe Left", NULL, 0},
  {"Strafe Right", NULL, 0},
  {"Jump", NULL, 0},
  {"Crouch", NULL, 0},
  {"Use / Open", NULL, 0},
  {"Primary Fire", NULL, 0},
  {"Secondary Fire", NULL, 0},
  {"Strafe", NULL, 0},
  {"Run", NULL, 0},
  {"Slow", NULL, 0},
  {"View Up", NULL, 0},
  {"View Down", NULL, 0},
  {"Center View", NULL, 0}
};

static char sms_bind[(34 + 1) * 13];
static menuitem ms_bind[] = {
  {"Mouse setup", NULL, 13},
  {"Turn Left / Right", NULL, 0},
  /* Forward / Backward */
  {"Move / Look Up/Down", NULL, 0}, /* v1.4 */
  {"Forward", NULL, 0},
  {"Backward", NULL, 0},
  {"Jump", NULL, 0},
  {"Crouch", NULL, 0},
  {"Use / Open", NULL, 0},
  {"Primary Fire", NULL, 0},
  {"Secondary Fire", NULL, 0}, /* v1.4 */
  {"Strafe", NULL, 0},
  {"Run", NULL, 0},
  {"Slow", NULL, 0},
  {"Center View", NULL, 0}
};

static menuitem ms_list[] = {
  {"Choose button or axis", NULL, 6},
  {"<None>", NULL, 0},
  {"X Axis", NULL, 0},
  {"Y Axis", NULL, 0},
  {"Left Button", NULL, 0},
  {"Right Button", NULL, 0},
  {"Middle Button", NULL, 0}
};

static char sjs_bind[(42 + 1) * 14];
static menuitem js_bind[] = {
  {"Controller setup", NULL, 15},
  {"Controller type", NULL, ITEM_HILITE},
  {"Enable Hat", NULL, ITEM_CHKBOX},
  {"Turn Left / Right", NULL, 0},
  {"Forward / Backward", NULL, 0},
  {"Strafe Left / Right", NULL, 0},
  {"Jump", NULL, 0},
  {"Crouch", NULL, 0},
  {"Use / Open", NULL, 0},
  {"Primary Fire", NULL, 0},
  {"Secondary Fire", NULL, 0},
  {"Strafe", NULL, 0},
  {"Run", NULL, 0},
  {"Slow", NULL, 0},
  {"View Up / Down", NULL, 0},
  {"Center View", NULL, 0}
};

static menuitem js_list[] = {
  {"Choose button or axis", NULL, 17},
  {"<None>", NULL, 0},
  {"X Axis", NULL, 0},
  {"Y Axis", NULL, 0},
  {"Button 1", NULL, 0},
  {"Button 2", NULL, 0},
  {"Button 3", NULL, 0},
  {"Button 4", NULL, 0},
  {"Button 1 Single Click", NULL, 0},
  {"Button 2 Single Click", NULL, 0},
  {"Button 3 Single Click", NULL, 0},
  {"Button 4 Single Click", NULL, 0},
  {"Button 1 Double Click", NULL, 0},
  {"Button 2 Double Click", NULL, 0},
  {"Button 3 Double Click", NULL, 0},
  {"Button 4 Double Click", NULL, 0},
  {"Hat X Axis", NULL, 0},
  {"Hat Y Axis", NULL, 0}
};

static menuitem js_type[] = {
  {"Choose controller type", NULL, 6},
  {"<None>", NULL, 0},
  {"Standard", NULL, 0},
  /* +3,4 buttons */
  {"Standard 4 Button", NULL, 0},
  {"Gravis Game Pad", NULL, 0},
  /* +hats */
  {"ThrustMaster FCS", NULL, 0},
  {"CH Flightstick Pro", NULL, 0}
};

static menuitem menusize[] = {
  {"Select screen size", NULL, 10},
  {"192x100", NULL, 0},
  {"224x110", NULL, 0},
  {"240x120", NULL, 0},
  {"256x130", NULL, 0},
  {"288x140", NULL, 0},
  {"304x144", NULL, 0},
  {"320x120", NULL, 0},
  {"320x160", NULL, 0},
  {"320x172", NULL, 0},
  {"320x200", NULL, 0}
};

static menuitem jedictrl[] = {
  {"Controls setup", "Current Filename ( JEDI.CFG )", 13}, /* v1.4 */
  {"Keyboard setup", "Setup key binding definitions.", ITEM_HILITE},
  {"Mouse setup", "Setup mouse button and axis definitions.", 0},
  {"Controller setup", "Setup game controller button and axis definitions.", 0},
  {"Disable keyboard", "Disables keyboard configuration, except reserved menu, map, weapons, etc. keys.", ITEM_CHKBOX},
  {"Disable mouse", "Disables mouse configuration, except cursor in menus.", ITEM_CHKBOX},
  {"Disable controller", "Disables controller configuration.", ITEM_CHKBOX},
  {"", NULL, ITEM_GRAYED},
  {"Select channels number: 8", "Allows to set number of simultaneous digital sound channels that can be played.", 0},
  {"Select screen size: 320x200", "System Speed setup: the smaller the size, the higher performance and vice versa.", 0},
  {"Enable weapon automount", "Automatically uses the next best weapon when the current weapon is empty.", ITEM_CHKBOX},
  {"", NULL, ITEM_GRAYED},
  {"Save changes to file", "Save changes to the configuration file.", 0},
  {"Do not save any changes", "Return without saving anything.", 0}
};

static menuitem menuload[] = {
  {"Invalid or absent configuration file!", NULL, 2 | MENU_DLGBOX},
  {"Use default", "Use default configuration for everything.", 0},
  {"Return back", "Return back to main menu.", 0}
};

static menuitem menusave[] = {
  {"Save changes to configuration?", NULL, 2 | MENU_DLGBOX},
  {"Yes", "Save all changes to file.", 0},
  {"No",  "Do not save any changes.", 0}
};

static menuitem menuchan[] = {
  {"Select channels number", NULL, 5},
  {"4 Digital Channels", NULL, 0},
  {"5 Digital Channels", NULL, 0},
  {"6 Digital Channels", NULL, 0},
  {"7 Digital Channels", NULL, 0},
  {"8 Digital Channels", NULL, 0}
};

static char strJediCfg[] = "JEDI.CFG";

/* this will add spaces to the menu items
   so the submenu item names can be added */
void FillMenu(menuitem *menu, char *s, int l) {
int i, j;
char *t;
  for (i = 0; i < (menu[0].flag & MENU_ITEMSK); i++) {
    /* skip Enable Hat item in joystick menu */
    if (!(menu[1 + i].flag & (ITEM_CHKBOX | ITEM_RADBOX))) {
      t = menu[1 + i].item;
      menu[1 + i].item = s;
      for (j = 0; j < l; j++) {
        *s = *t ? *t : ' ';
        t += *t ? 1 : 0;
        s++;
      }
      *s = 0;
      s++;
    }
  }
}

void SetItemData(menuitem *item, char *s, BYTE v, int p) {
  StrOverlap(&item->item[p ? p : 21], s);
  SetMenuData(*item, v);
}

int ShowInfo(char *s, char *h, PMENUPROC mp) {
menuitem item;
  item.item = s;
  item.hint = h;
  item.flag = mp ? 0 : 1;
  return(DrawInfo(&item, mp));
}

int GetKeyProc(int item) {
  return((!item) ? GetScanCode() : 0);
}

int KBMenu(int item) {
char *s, b[60];
WORD l;
  do {
    l = ShowInfo("Press a Key", "<Esc> to remove", GetKeyProc);
    s = GetNameByScanCode(l);
    /* Escape pressed - use <None> as key */
    if (l == 1) { l = 0; }
    /* Escape or <None> */
    if (!l) { break; }
    /* check if that key allowed to assign */
    if (IsReservedKey(l)) {
      StrMerge(b, 60, "ERROR: <", s, "> already in use as reserved game key.", NULL);
      ShowInfo(b, NULL, NULL);
      l = 0;
    }
  } while (!l);
  s = GetNameByScanCode(l);
  SetItemData(&kb_bind[item], s, l, 0);
  return(1);
}

int MSMenu(int item) {
int i;
  DefMenuItem(ms_list, GetMenuData(ms_bind[item]));
  i = DrawMenu(ms_list, NULL);
  if (i) {
    SetItemData(&ms_bind[item], ms_list[i].item, i - 1, 0);
  }
  return(1);
}

int JSMenu(int item) {
int i;
  /* joystick type */
  if (item == 1) {
    DefMenuItem(js_type, GetMenuData(js_bind[item]));
    i = DrawMenu(js_type, NULL);
    if (i) {
      SetItemData(&js_bind[item], js_type[i].item, i - 1, 0);
      /* fallback - init hat and buttons */
      item = 2;
    }
  }
  /* hat on/off and buttons */
  if (item == 2) {
    /* buttons 3 and 4 available */
    i = (GetMenuData(js_bind[1]) < 2) ? 1 : 0;
    SET_GRAY(js_list[6], i);
    SET_GRAY(js_list[7], i);
    SET_GRAY(js_list[10], i);
    SET_GRAY(js_list[11], i);
    SET_GRAY(js_list[14], i);
    SET_GRAY(js_list[15], i);
    /* hats available */
    i = (GetMenuData(js_bind[1]) < 4) ? 1 : 0;
    SET_GRAY(js_bind[item], i);
    if (i) {
      SET_LITE(js_bind[item], 0);
    }
    /* hats enabled */
    i = (js_bind[item].flag & ITEM_HILITE) ? 0 : 1;
    SET_GRAY(js_list[16], i);
    SET_GRAY(js_list[17], i);
    /* special case to update menu items */
    item = 0;
  }
  /* bindings */
  if (item >= 3) {
    DefMenuItem(js_list, GetMenuData(js_bind[item]));
    i = DrawMenu(js_list, NULL);
    if (i) {
      SetItemData(&js_bind[item], js_list[i].item, i - 1, 0);
    }
  }
  /* update this menu items if needed */
  return(item ? 1 : -1);
}

int EditMenu(int item) {
int i;
  switch (item) {
    case 1:
      if (DrawTest(kb_bind, KBMenu)) {
        SetMenuData(jedictrl[1], 1); /* v1.4 */
      }
      break;
    case 2:
      if (DrawTest(ms_bind, MSMenu)) {
        SetMenuData(jedictrl[1], 1); /* v1.4 */
      }
      break;
    case 3:
      if (DrawTest(js_bind, JSMenu)) {
        SetMenuData(jedictrl[1], 1); /* v1.4 */
      }
      break;
    case 4:
    case 5:
    case 6:
    case 10:
      SetMenuData(jedictrl[1], 1); /* v1.4 */
      break;
    case 8:
      DefMenuItem(menuchan, GetMenuData(jedictrl[item]) - 4);
      i = DrawMenu(menuchan, NULL);
      if (i) {
        SetItemData(&jedictrl[item], menuchan[i].item, i - 1 + 4, 24);
        SetMenuData(jedictrl[1], 1); /* v1.4 */
      }
      break;
    case 9:
      DefMenuItem(menusize, GetMenuData(jedictrl[item]));
      i = DrawMenu(menusize, NULL);
      if (i) {
        SetItemData(&jedictrl[item], menusize[i].item, i - 1, 20);
        SetMenuData(jedictrl[1], 1); /* v1.4 */
      }
      break;
  }
  return((item + 1) < (jedictrl[0].flag & MENU_ITEMSK));
}

int LoadJediConfig(int flag) {
int i, k;
int fl;
  /* init menus */
  FillMenu(kb_bind, skb_bind, 32);
  FillMenu(ms_bind, sms_bind, 34);
  FillMenu(js_bind, sjs_bind, 42);
  /* v1.4 - no flags */
  if (!flag) {
    /* fill with zeros */
    memset(&dfconfig, 0, sizeof(dfconfig));
    /* check that file exists */
    if (FileExists(strJediCfg, sizeof(dfconfig))) {
      /* read file */
      fl = open(strJediCfg, O_RDONLY | O_BINARY);
      if (fl != -1) {
        read(fl, &dfconfig, sizeof(dfconfig));
        close(fl);
      }
    }
    /* check magic numbers */
    fl = (
      (dfconfig.mark[0] == 'C') && (dfconfig.mark[1] == 'F') &&
      (dfconfig.mark[2] == 'G') && (dfconfig.vers == 100) &&
      (dfconfig.len1 == 60) && (dfconfig.len2 == 37) &&
      (dfconfig.len3 == 84) && (dfconfig.len4 == 57)
    );
    /* some error - ask for default settings */
    if (!fl) {
      if (DrawMenu(menuload, 0) == 1) {
        memcpy(&dfconfig, defcfg, sizeof(dfconfig));
        fl = 1;
        SetMenuData(jedictrl[1], 1); /* v1.4 */
      }
    }
  } else {
    /* load default modern configuration */
    memcpy(&dfconfig, mdrcfg, sizeof(dfconfig));
    fl = 1;
  }
  /* everything seems ok */
  if (fl) {
    /* if mouse sensitivity zero - mouse disabled */
    dfconfig.main.bAllowMS = dfconfig.main.iMouseSensitivity ? DF_TRUE : DF_FALSE;
    /* if no joystick type selected - joystick disabled */
    dfconfig.main.bAllowJS = dfconfig.list[0] ? DF_TRUE : DF_FALSE;
    /* state flags */
    SET_LITE(jedictrl[4], dfconfig.main.bAllowKB == DF_FALSE);
    SET_LITE(jedictrl[5], dfconfig.main.bAllowMS == DF_FALSE);
    SET_LITE(jedictrl[6], dfconfig.main.bAllowJS == DF_FALSE);
    /* keyboard */
    for (i = 0; i < LIST_LEN(kbidx); i++) {
      k = dfconfig.ctrl.kb[kbidx[i]];
      SetItemData(&kb_bind[1 + i], GetNameByScanCode(k), k, 0);
    }
    /* mouse */
    for (i = 0; i < LIST_LEN(msidx); i++) {
      k = dfconfig.ctrl.ms[msidx[i]];
      /* special case: only first item can be 1 or 2 */
      k -= (i && (k > 2)) ? 2 : 0;
      SetItemData(&ms_bind[1 + i], ms_list[1 + k].item, k, 0);
    }
    /* joystick */
    for (i = 0; i < LIST_LEN(jsidx); i++) {
      k = dfconfig.ctrl.js[jsidx[i]];
      SetItemData(&js_bind[3 + i], js_list[1 + k].item, k, 0);
    }
    /* joystick type */
    i = jstom[((BYTE) dfconfig.list[0]) & 0x07];
    SetItemData(&js_bind[1], js_type[i + 1].item, i, 0);
    /* hats allowed */
    if (i < 4) {
      SET_GRAY(js_bind[2], 1);
      i = 0;
    } else {
      /* 4,5 => 1,2 */
      i = (BYTE) dfconfig.list[i - 3];
    }
    SET_LITE(js_bind[2], i);
    /* init hats */
    JSMenu(2);
    /* digital channels */
    if (dfconfig.main.iChannelsNumber < 4) { dfconfig.main.iChannelsNumber = 4; }
    if (dfconfig.main.iChannelsNumber > 8) { dfconfig.main.iChannelsNumber = 8; }
    SetItemData(&jedictrl[8], menuchan[1 + ((int) dfconfig.main.iChannelsNumber) - 4].item, dfconfig.main.iChannelsNumber, 24);
    /* screen size - reverse for top (lower) to bottom (higher) resolution list */
    if (dfconfig.main.iScreenSize > 9) { dfconfig.main.iScreenSize = 9; }
    SetItemData(&jedictrl[9], menusize[1 + 9 - ((int) dfconfig.main.iScreenSize)].item, 9 - dfconfig.main.iScreenSize, 20);
    /* Weapon AutoMount
       SETUP.EXE so bugged - it's only set 1st byte out of 4 bytes value,
       but the game itself checks this value as whole!
    */
    SET_LITE(jedictrl[10], dfconfig.main.bWeaponAutoMount & 0xFF);
  }
  return(fl);
}

void SaveJediConfig(void) {
int i, k;
int fl;
  /* state flags */
  dfconfig.main.bAllowKB = (jedictrl[4].flag & ITEM_HILITE) ? DF_FALSE : DF_TRUE;
  dfconfig.main.bAllowMS = (jedictrl[5].flag & ITEM_HILITE) ? DF_FALSE : DF_TRUE;
  dfconfig.main.bAllowJS = (jedictrl[6].flag & ITEM_HILITE) ? DF_FALSE : DF_TRUE;
  /* if mouse enabled - raise mouse sensitivity from zero */
  if (dfconfig.main.bAllowMS) {
    /* try to keep selected sensitivity instead of overwriting with default one */
    if (!dfconfig.main.iMouseSensitivity) {
      dfconfig.main.iMouseSensitivity = 4; /* default value */
    }
  } else {
    /* unfortunately sensitivity required to be zero to complete disable mouse */
    dfconfig.main.iMouseSensitivity = 0;
  }
  /* if joystick enabled it's type must be different from none */
  i = GetMenuData(js_bind[1]);
  if (dfconfig.main.bAllowJS) {
    /* if no joystick type selected - disabled joystick regardless of setting */
    dfconfig.main.bAllowJS = i ? DF_TRUE : DF_FALSE;
  } else {
    /* global enable / disable settings takes priority - remove joystick type */
    SetItemData(&js_bind[1], NULL, 0, 0);
  }
  /* keyboard */
  for (i = 0; i < LIST_LEN(kbidx); i++) {
    dfconfig.ctrl.kb[kbidx[i]] = GetMenuData(kb_bind[1 + i]);
  }
  /* keyboard map */
  for (i = 0; i < LIST_LEN(kbmap); i++) {
    dfconfig.ctrl.kb[kbmap[i][0]] = dfconfig.ctrl.kb[kbmap[i][1]];
  }
  /* mouse */
  for (i = 0; i < LIST_LEN(msidx); i++) {
    k = GetMenuData(ms_bind[1 + i]);
    /* special case: only first item can be 1 or 2 */
    k += (i && k) ? 2 : 0;
    dfconfig.ctrl.ms[msidx[i]] = k;
  }
  /* joystick */
  for (i = 0; i < LIST_LEN(jsidx); i++) {
    dfconfig.ctrl.js[jsidx[i]] = GetMenuData(js_bind[3 + i]);
  }
  /* joystick map */
  for (i = 0; i < LIST_LEN(jsmap); i++) {
    dfconfig.ctrl.js[jsmap[i][0]] = dfconfig.ctrl.js[jsmap[i][1]];
  }
  /* joystick type */
  i = GetMenuData(js_bind[1]);
  /* if different type was selected */
  if (dfconfig.list[0] != jstof[i]) {
    dfconfig.list[0] = jstof[i];
    /* joystick not calibrated or already calibrated if <None> selected
       this will bring Joystick Calibration menu on startup
       but only if there are any actual joystick detected in the system
       (note that /C0 commandline option will force to skip that menu) */
    dfconfig.list[20] = dfconfig.list[0] ? DF_FALSE : DF_TRUE;
  }
  /* hats enabled */
  k = (js_bind[2].flag & ITEM_HILITE) ? 1 : 0;
  /* hats allowed */
  dfconfig.list[1] = ((i == 4) && k) ? 4 : 0;
  dfconfig.list[2] = ((i == 5) && k) ? 1 : 0;
  /* channels */
  dfconfig.main.iChannelsNumber = GetMenuData(jedictrl[8]);
  /* screen size - reverse back */
  dfconfig.main.iScreenSize = 9 - GetMenuData(jedictrl[9]);
  /* this MUST be -1 (4 bytes!) or 0 otherwise the game can't work with it */
  dfconfig.main.bWeaponAutoMount = (jedictrl[10].flag & ITEM_HILITE) ? DF_TRUE : DF_FALSE;
  /* create config file */
  fl = open(strJediCfg, O_RDWR | O_BINARY | O_CREAT | O_TRUNC, S_IREAD | S_IWRITE);
  if (fl != -1) {
    write(fl, &dfconfig, sizeof(dfconfig));
    close(fl);
    /* check that file fully saved */
    fl = FileExists(strJediCfg, sizeof(dfconfig)) ? fl : -1;
  }
  /* something went wrong */
  if (fl == -1) {
    ShowInfo("Error saving JEDI.CFG! Check it's not read-only and enough free disk space.", NULL, NULL);
  }
}

void EditJediConfig(int flag) {
int i;
  SetMenuData(jedictrl[1], 0); /* v1.4 ask about changes */
  if (LoadJediConfig(0)) {
    SET_GRAY(ms_bind[9], flag ? 0 : 1); /* v1.4 - [Fire Secondary] on mouse must be enabled */
    i = DrawMenu(jedictrl, EditMenu);
    /* save if needed */
    if (
      /* v1.4 something changed */
      GetMenuData(jedictrl[1]) && (
        /* save selected */
        ((i + 1) == (jedictrl[0].flag & MENU_ITEMSK)) ||
        /* or ask for saving changes */
        ((!i) && (DrawMenu(menusave, NULL) == 1))
      )
    ) {
      SaveJediConfig();
    }
  }
}

/* v1.4 */
void MakeModernConfig(void) {
  LoadJediConfig(1);
  SaveJediConfig();
}
