#ifdef _WIN32

#include "WIN32MOD.H"
#include "XSCOMMON.H"

int close(int handle) {
  return(CloseHandle((HANDLE) handle) ? 0 : -1);
}

long lseek(int handle, long offset, int fromwhere) {
  return(SetFilePointer((HANDLE) handle, offset, NULL, fromwhere));
}

int open(const char *path, int access, ...) {
DWORD ac, cd;
  ac = 0;
  cd = OPEN_EXISTING;
  switch (access & 3) {
    case 0:
      ac = GENERIC_READ;
      break;
    case 1:
      ac = GENERIC_WRITE;
      break;
    case 2:
      ac = GENERIC_READ | GENERIC_WRITE;
      break;
  }
  if (access & O_CREAT) { cd = CREATE_NEW; }
  if (access & O_TRUNC) { cd = TRUNCATE_EXISTING; }
  if ((access & (O_CREAT | O_TRUNC)) == (O_CREAT | O_TRUNC)) { cd = CREATE_ALWAYS; }
  return((int) CreateFile(path, ac, 0, NULL, cd, FILE_ATTRIBUTE_NORMAL, 0));
}

int read(int handle, void *buf, unsigned int len) {
DWORD dw;
  dw = 0;
  return(ReadFile((HANDLE) handle, buf, len, &dw, NULL) ? ((int) dw) : -1);
}

long tell(int handle) {
  return(SetFilePointer((HANDLE) handle, 0, NULL, FILE_CURRENT));
}

int write(int handle, const void *buf, unsigned int len) {
DWORD dw;
  dw = 0;
  return(WriteFile((HANDLE) handle, buf, len, &dw, NULL) ? ((int) dw) : -1);
}

int setftime(int handle, struct ftime *ftimep) {
  if (!ftimep) { handle = -1; }
  if (handle != -1) {
    handle = SetFileTime(
      (HANDLE) handle,
      &ftimep->ftCreationTime,
      &ftimep->ftLastAccessTime,
      &ftimep->ftLastWriteTime
    ) ? 0 : -1;
  }
  return(handle);
}

int getftime(int handle, struct ftime *ftimep) {
  if (!ftimep) { handle = -1; }
  if (handle != -1) {
    handle = GetFileTime(
      (HANDLE) handle,
      &ftimep->ftCreationTime,
      &ftimep->ftLastAccessTime,
      &ftimep->ftLastWriteTime
    ) ? 0 : -1;
  }
  return(handle);
}

int bioskey(int cmd) {
INPUT_RECORD ir;
DWORD dw, cm;
HANDLE h;
  h = GetStdHandle(STD_INPUT_HANDLE);
  if (h) {
    GetConsoleMode(h, &cm);
    SetConsoleMode(h, cm & ~(ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT | ENABLE_WINDOW_INPUT | ENABLE_MOUSE_INPUT));
    FlushConsoleInputBuffer(h);
    if (!cmd) {
      do {
        WaitForSingleObject(h, INFINITE);
      } while (!(
        (ReadConsoleInput(h, &ir, 1, &dw)) && (dw == 1) &&
        (ir.EventType == KEY_EVENT) && (ir.Event.KeyEvent.bKeyDown)
      ));
      cmd = (ir.Event.KeyEvent.wVirtualKeyCode << 8);
    } else {
      cmd = (
        (PeekConsoleInput(h, &ir, 1, &dw)) && (dw == 1) &&
        (ir.EventType == KEY_EVENT) && (ir.Event.KeyEvent.bKeyDown)
      ) ? (ir.Event.KeyEvent.wVirtualKeyCode << 8) : 0;
    }
    FlushConsoleInputBuffer(h);
    SetConsoleMode(h, cm);
  } else {
    cmd = VK_ESCAPE;
  }
  return(cmd);
}

/* Windows 2k / XP */
#pragma pack(push, 1)
typedef struct {
  DWORD nFont;
  COORD dwFontSize;
} CONSOLE_FONT_INFO;
#pragma pack(pop)

typedef DWORD (WINAPI *LPGETNUMBEROFCONSOLEFONTS)(void);
typedef LONG  (WINAPI *LPGETCONSOLEFONTINFO)(HANDLE hConsoleOutput, BOOL bMaximumWindow, DWORD nLength, CONSOLE_FONT_INFO *lpConsoleFontInfo);
typedef COORD (WINAPI *LPGETCONSOLEFONTSIZE)(HANDLE hConsoleOutput, DWORD nFont);
typedef BOOL  (WINAPI *LPSETCONSOLEFONT)(HANDLE hConsoleOutput, DWORD nFont);

/* Windows Vista / 7 / 8 / 10 */
#pragma pack(push, 1)
typedef struct {
  ULONG cbSize;
  DWORD nFont;
  COORD dwFontSize;
  UINT  FontFamily;
  UINT  FontWeight;
  WCHAR FaceName[LF_FACESIZE];
} CONSOLE_FONT_INFOEX;
#pragma pack(pop)

#ifndef CONSOLE_FULLSCREEN_MODE
#define CONSOLE_FULLSCREEN_MODE 1
#endif

typedef BOOL (WINAPI *LPGETCURRENTCONSOLEFONTEX)(HANDLE hConsoleOutput, BOOL bMaximumWindow, CONSOLE_FONT_INFOEX *lpConsoleCurrentFontEx);
typedef BOOL (WINAPI *LPSETCURRENTCONSOLEFONTEX)(HANDLE hConsoleOutput, BOOL bMaximumWindow, CONSOLE_FONT_INFOEX *lpConsoleCurrentFontEx);

void ChangeConsoleFontSize(HANDLE h) {
LPGETNUMBEROFCONSOLEFONTS GetNumberOfConsoleFonts;
LPGETCONSOLEFONTINFO GetConsoleFontInfo;
LPGETCONSOLEFONTSIZE GetConsoleFontSize;
LPSETCONSOLEFONT SetConsoleFont;
LPGETCURRENTCONSOLEFONTEX GetCurrentConsoleFontEx;
LPSETCURRENTCONSOLEFONTEX SetCurrentConsoleFontEx;
CONSOLE_FONT_INFOEX cfi;
CONSOLE_FONT_INFO *f;
HMODULE hm;
DWORD i, k;
  hm = GetModuleHandle(TEXT("KERNEL32.dll"));
  if (hm && h) {
    /*
      SetConsoleDisplayMode(h, ?, &?);
      https://docs.microsoft.com/en-us/windows/console/setconsoledisplaymode
    */
    /* Windows 2k / XP */
    GetNumberOfConsoleFonts = (LPGETNUMBEROFCONSOLEFONTS) GetProcAddress(hm, "GetNumberOfConsoleFonts");
    GetConsoleFontInfo = (LPGETCONSOLEFONTINFO) GetProcAddress(hm, "GetConsoleFontInfo");
    GetConsoleFontSize = (LPGETCONSOLEFONTSIZE) GetProcAddress(hm, "GetConsoleFontSize");
    SetConsoleFont = (LPSETCONSOLEFONT) GetProcAddress(hm, "SetConsoleFont");
    GetCurrentConsoleFontEx = (LPGETCURRENTCONSOLEFONTEX) GetProcAddress(hm, "GetCurrentConsoleFontEx");
    SetCurrentConsoleFontEx = (LPSETCURRENTCONSOLEFONTEX) GetProcAddress(hm, "SetCurrentConsoleFontEx");
    if (GetNumberOfConsoleFonts && GetConsoleFontInfo && GetConsoleFontSize && SetConsoleFont) {
      k = GetNumberOfConsoleFonts();
      if (k) {
        f = (CONSOLE_FONT_INFO *) LocalAlloc(LMEM_FIXED, sizeof(f[0]) * k);
        if (f) {
          GetConsoleFontInfo(h, FALSE, k, f);
          for (i = 0; i < k; i++) {
            f[i].dwFontSize = GetConsoleFontSize(h, f[i].nFont);
            if ((f[i].dwFontSize.X == 10) && (f[i].dwFontSize.Y == 18)) {
              SetConsoleFont(h, f[i].nFont);
              break;
            }
          }
          LocalFree(f);
        }
      }
    }
    /* Windows Vista / 7 / 8 / 10 */
    if (GetCurrentConsoleFontEx && SetCurrentConsoleFontEx) {
      ZeroMemory(&cfi, sizeof(cfi));
      cfi.cbSize = sizeof(cfi);
      GetCurrentConsoleFontEx(h, FALSE, &cfi);
      cfi.dwFontSize.X = 10;
      cfi.dwFontSize.Y = 18;
      SetCurrentConsoleFontEx(h, FALSE, &cfi);
    }
  }
}

/* old GCC headers */
#ifndef KEY_WOW64_64KEY
#define KEY_WOW64_64KEY (0x0100)
#endif

static char *strblock[][3] = {
  /* Current folder */
  {
    "",
    "",
    ""
  },
  /* GOG Galaxy */
  {
    "SOFTWARE\\GOG.com\\Games\\1421404433",
    "path",
    "\\"
  },
  /* Steam */
  {
    "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\Steam App 32400",
    "InstallLocation",
    "\\Game\\"
  }
};

void RegReadPath(char *path, char *name, int b64, char *buff, DWORD size) {
DWORD dt, dw;
HKEY hr;
  if (buff && size) {
    *buff = 0;
    if (path && name && (*path) && (*name) && (
      RegOpenKeyEx(
        HKEY_LOCAL_MACHINE, path, 0, KEY_QUERY_VALUE | (b64 ? KEY_WOW64_64KEY : 0), &hr
      ) == ERROR_SUCCESS)
    ) {
      dt = REG_SZ;
      dw = size;
      if (RegQueryValueEx(hr, name, NULL, &dt, (BYTE *) buff, &dw) != ERROR_SUCCESS) {
        dw = 0;
      } else {
        /* RegQueryValueEx() for REG_SZ type MAY NOT terminate string with the zero */
        if (dw && buff[dw - 1]) {
          dw -= (dw < size) ? 0 : 1;
        }
      }
      /* make always null terminated string */
      buff[dw] = 0;
      RegCloseKey(hr);
    }
  }
}

int FindGameFile(char *name, unsigned long size, int flag) {
char path[1024], disk[4];
int k, r, fl;
DWORD dw, ld;
  r = 0;
  for (k = 0; k < 3; k++) {
    *path = 0;
    /* read x32 */
    RegReadPath(strblock[k][0], strblock[k][1], 0, path, 1000);
    /* read x64 */
    if (!*path) {
      RegReadPath(strblock[k][0], strblock[k][1], 1, path, 1000);
    }
    /* current folder or got something */
    if ((!k) || (*path)) {
      dw = strlen(path);
      while (dw--) {
        if ((path[dw] == '\\') || (path[dw] == '/')) {
          path[dw] = 0;
        } else {
          break;
        }
      }
      StrMerge(path, 1024, path, strblock[k][2], name, NULL);
      /* mostly for Steam if folder moved */
      if (path[1] == ':') {
        dw = SetErrorMode(SEM_FAILCRITICALERRORS);
        ld = GetLogicalDrives();
        disk[0] = 'A';
        disk[1] = ':';
        disk[2] = '\\';
        disk[3] = 0;
        while (ld) {
          if (ld & 1) {
            if (GetDriveType(disk) == DRIVE_FIXED) {
              *path = *disk;
              if (FileExists(path, size)) {
                break;
              }
            }
          }
          disk[0]++;
          ld >>= 1;
        }
        SetErrorMode(dw);
      }
      /* file exist */
      if (FileExists(path, size)) {
        /* file exists */
        r |= 1 << k;
        /* check access rights (run as Administrator for Program Files folder required) */
        fl = open(path, O_RDWR | O_BINARY);
        if (fl != -1) {
          close(fl);
          /* set writeable mask */
          r |= 1 << (3 + k);
        }
      }
      /* break if flag */
      if (flag && ((r & flag) == flag)) {
        dw = strlen(path) - strlen(name);
        /* change working folder */
        if (dw) {
          path[dw] = 0;
          SetCurrentDirectory(path);
        }
        break;
      }
    }
  }
  return(r);
}

#endif
